<?php

namespace Drupal\rijksvideo\EventSubscriber;

use Drupal\Component\Utility\Bytes;
use Drupal\file\Entity\File;
use Drupal\rijksvideo\Event\RijksvideoEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Drupal\Core\File\FileSystemInterface;

/**
 * Enrich the creation of a new media entity.
 */
class RijksvideoEventSubscriber implements EventSubscriberInterface {

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[RijksvideoEvent::RIJKSVIDEO_MEDIA_PRESAVE][] = ['fillMetaData'];
    return $events;
  }

  /**
   * Enrich the creation of a new media entity.
   *
   * @param \Drupal\rijksvideo\Event\RijksvideoEvent $event
   *   The Rijksvideo event.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function fillMetaData(RijksvideoEvent $event) {
    $entity = $event->getEntity();

    if ($entity->bundle() !== "rijksvideo") {
      return;
    }

    $fid = $entity->get('field_rijksvideo_file')->target_id;
    $file = File::load($fid);

    $xmlData = file_get_contents($file->getFileUri());
    $result = new \SimpleXMLElement($xmlData);

    $rijksvideo = [];
    $title = (string) $result->title;
    $entity->setName($title);

    foreach ($result->files->children() as $file) {
      $url = (string) $file;
      // Get attributes from file.
      $attributes = (array) $file->attributes();
      $attributes = $attributes['@attributes'];

      switch ($attributes['type']) {
        case 'still_m':
          $ext = pathinfo($url, PATHINFO_EXTENSION);
          $thumb = $this->rijksvideoFetchFile($url, $title, '_thumb.' . $ext);
          $entity->set('field_rijksvideo_thumbnail', $thumb);
          $entity->thumbnail = $thumb;
          break;

        case 'subtitle_srt':
          $entity->set('field_rijksvideo_caption', $this->rijksvideoFetchFile($url, $title, '_subtitle.srt'));
          break;

        case 'text':
          $entity->set('field_rijksvideo_transcription', $this->rijksvideoFetchTranscription($url));
          break;

        case 'audio_mp3':
          $entity->set('field_rijksvideo_audio', $this->normalizeUrl($url));
          break;

        // Video files.
        case 'web_webm':
        case 'web_mp4':
        case 'web_wmv':
        case 'web_flv':
        case 'web_hd':
        case 'web_ogv':
          $rijksvideo[$attributes['type']]['url'] = $this->normalizeUrl($url);
          $rijksvideo[$attributes['type']]['size'] = $this->formatBytes(Bytes::toInt($attributes['size']));
          $rijksvideo[$attributes['type']]['type'] = strtoupper(str_replace(['web_', 'hd'], ['', 'mp4 hd'], $attributes['type']));
          break;
      }
    }

    $entity->set('field_rijksvideo_duration', $result->duration);
    $entity->set('field_rijksvideo_sources', json_encode($rijksvideo));

    $event->setEntity($entity);
  }

  /**
   * Fetch file or subittles.
   *
   * @param string $url
   *   The URL that needs to be fetched.
   * @param string $title
   *   The title of the file.
   * @param string $type
   *   The type of file.
   *
   * @return mixed
   *   Return either the file or subtitles.
   */
  protected function rijksvideoFetchFile($url, $title, $type) {
    if (empty($url) || !$this->urlExists($url)) {
      return FALSE;
    }

    $url = $this->normalizeUrl($url);
    $title = strtolower(str_replace(' ', '_', $title));
    $title = str_replace([" ", '"', "'", "&", "/", "\\", "?", "#"], "", $title);
    $path = 'public://rv_data_' . $title . $type;

    if ($type != '_subtitle.srt') {
      return system_retrieve_file($url, $path, TRUE, FileSystemInterface::EXISTS_RENAME);
    }

    // File is a subtitle, download its content.
    $response = \Drupal::httpClient()->get($url, ['headers' => ['Accept' => 'text/plain']]);
    $file_content = (string) $response->getBody();

    // Convert SRT files to VTT.
    $file = $this->convertSrtToVtt($file_content);

    // Save the new subtitle file.
    $subtitles = file_save_data($file, str_replace('.srt', '.vtt', $path));

    return $subtitles;
  }

  /**
   * Check whether a URL exists.
   *
   * @param string $url
   *   The URL that needs to be checked.
   *
   * @return bool
   *   Returns if the URL exists
   */
  protected function urlExists($url) {
    $url = $this->normalizeUrl($url);

    $ch = curl_init($url);

    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, FALSE);
    curl_setopt($ch, CURLOPT_NOBODY, TRUE);
    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 5);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    $result = curl_exec($ch);

    if ($result !== FALSE) {
      $httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
      curl_close($ch);
      if ($httpCode == 200) {
        return TRUE;
      }
    }

    \Drupal::logger('rijksvideo')->error("Could not download file: " . $url . " - It probably doesn't exist!");

    return FALSE;
  }

  /**
   * Normalize a URL.
   *
   * @param string $url
   *   The URL that needs to be normalized.
   *
   * @return mixed|string
   *   The normalized URL.
   */
  protected function normalizeUrl($url) {
    $url = trim($url);
    $url = str_replace(['http://', 'https://'], '', $url);
    $url = 'https://' . $url;

    return trim($url);
  }

  /**
   * Converts file's content (.srt) to vtt format.
   *
   * @param string $file_content
   *   Content of file that will be converted.
   *
   * @return array
   *   Internal format
   */
  protected function convertSrtToVtt($file_content) {
    if (strlen($file_content) < 10) {
      return $file_content;
    }

    // Array - where file content will be stored.
    $internal_format = [];
    $blocks = preg_split("#\n\s*\n#Uis", trim($file_content));

    foreach ($blocks as $block) {
      // Separate all block lines.
      $lines = explode("\n", $block);

      // One the second line there is start and end times.
      $times = explode(' --> ', $lines[1]);
      $internal_format[] = [
        'start' => static::srtTimeToInternal($times[0]),
        'end' => static::srtTimeToInternal($times[1]),
      // Get all the remaining lines from block (if multiple lines of text)
        'lines' => array_slice($lines, 2),
      ];
    }

    $file_content = "WEBVTT\r\n\r\n";

    foreach ($internal_format as $block) {
      $start = static::internalTimeToVtt($block['start']);
      $end = static::internalTimeToVtt($block['end']);
      $lines = implode("\n", $block['lines']);
      $file_content .= $start . ' --> ' . $end . "\r\n";
      $file_content .= $lines . "\r\n";
      $file_content .= "\r\n";
    }
    $file_content = trim($file_content);

    return $file_content;
  }

  /**
   * Convert .srt file format to internal time format (float in seconds)
   *
   * Example: 00:02:17,440 -> 137.44.
   *
   * @param string $srt_time
   *   The SRT time.
   *
   * @return float
   *   The internal time.
   */
  protected static function srtTimeToInternal($srt_time) {
    $parts = explode(',', $srt_time);
    $only_seconds = strtotime("1970-01-01 {$parts[0]} UTC");
    $milliseconds = (float) ('0.' . $parts[1]);
    $time = $only_seconds + $milliseconds;

    return $time;
  }

  /**
   * Convert internal time to SRT time.
   *
   * @param string $internal_time
   *   The internal time.
   *
   * @return string
   *   The SRT time.
   */
  protected static function internalTimeToVtt($internal_time) {
    // 1.23
    $parts = explode('.', $internal_time);
    // 1
    $whole = $parts[0];
    // 23
    $decimal = isset($parts[1]) ? substr($parts[1], 0, 3) : 0;
    $srt_time = gmdate("H:i:s", floor($whole)) . '.' . str_pad($decimal, 3, '0', STR_PAD_RIGHT);

    return $srt_time;
  }

  /**
   * Fetch transcription.
   *
   * @param string $url
   *   The URL that needs to be fetched.
   *
   * @return bool|string
   *   Returns the transcription.
   */
  protected function rijksvideoFetchTranscription($url) {
    if (!empty($url) && $this->urlExists($url)) {
      $url = $this->normalizeUrl($url);

      // File is a subtitle, download its content.
      $transcription = file_get_contents($url);
      $transcription = iconv(mb_detect_encoding($transcription, mb_detect_order(), TRUE), "UTF-8//TRANSLIT//IGNORE", $transcription);
      return $transcription;
    }
  }

  /**
   * @param $bytes
   * @param int $precision
   *
   * @return string
   */
  protected function formatBytes($bytes, $precision = 2) {
    // return $bytes;
    $units = ['B', 'KB', 'MB', 'GB', 'TB'];

    $bytes = max($bytes, 0);
    $pow = floor(($bytes ? log($bytes) : 0) / log(1024));
    $pow = min($pow, count($units) - 1);

    $bytes /= pow(1024, $pow);

    return round($bytes, $precision) . ' ' . $units[$pow];
  }

}
